# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from deluca.core import Agent
from deluca.core import field
from deluca.core import Obj


class PIDState(Obj):
    P: float = field(0.0, trainable=True)
    I: float = field(0.0, trainable=True)
    D: float = field(0.0, trainable=True)


class PID(Agent):
    K_P: float = field(0.0, trainable=True)
    K_I: float = field(0.0, trainable=True)
    K_D: float = field(0.0, trainable=True)
    RC: float = field(0.5, trainable=False)
    dt: float = field(0.03, trainable=False)
    decay: float = field(trainable=False)

    def init(self):
        return PIDState()

    def setup(self):
        self.decay = self.dt / (self.dt + self.RC)

    def __call__(self, state, obs):
        """Assume observation is err"""
        P = obs
        I = state.I + self.decay * (obs - state.I)
        D = state.D + self.decay * (obs - state.P - state.D)

        action = self.K_P * P + self.K_I * I + self.K_D * D

        return state.replace(P=P, I=I, D=D), action



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
TODO:
- jax.core.valid_jaxtype checks values, but not types, so we don't have a check
  that state is a jax type
- Explain why dynamics etc are not static methods (why pass self explicitly as
arg?)
- extracting state from the env object is a little messy

QUESTIONS:
- should env/agent be frozen?

"""
import dataclasses
import inspect
import os
import pickle
from abc import abstractmethod

import flax
import jax

import deluca


def save(obj, path):
    dirname = os.path.abspath(os.path.dirname(path))
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(path, "wb") as file:
        pickle.dump(obj, file)


def load(path):
    return pickle.load(open(path, "rb"))


def field(default=None, trainable=True, **kwargs):
    if "default_factory" not in kwargs:
        kwargs["default"] = default
    kwargs["pytree_node"] = trainable
    return flax.struct.field(**kwargs)


def update_signature(src, dst):
    if hasattr(src, "__doc__") and src.__doc__ is not None:
        dst.__doc__ = src.__doc__

    def wrapper(*args, **kwargs):
        return src(*args, **kwargs)

    wrapper.__signature__ = inspect.signature(src.__init__)
    return wrapper


class Obj:
    def __new__(cls, *args, **kwargs):
        """A true bastardization of __new__..."""

        def __setattr__(self, name, value):
            if self.__frozen__:
                raise dataclasses.FrozenInstanceError
            object.__setattr__(self, name, value)

        cls.__setattr__ = __setattr__
        obj = object.__new__(cls)
        object.__setattr__(obj, "__frozen__", False)
        obj.create = update_signature(cls.__init__, obj.create)

        return obj

    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        flax.struct.dataclass(cls)

    @classmethod
    def create(cls, *args, **kwargs):
        # NOTE: Oh boy, this is so janky
        dummy = cls(*args, **kwargs)
        dummy.setup()
        kwargs = {key: getattr(dummy, key) for key in dummy.__dataclass_fields__.keys()}

        obj = cls(**kwargs)

        for key in dummy.__dict__.keys():
            if not hasattr(obj, key):
                setattr(obj, key, dummy.__dict__[key])
        obj.__frozen__ = True

        return obj

    @classmethod
    def unflatten(cls, treedef, leaves):
        """Expost a default unflatten method"""
        return jax.tree_util.tree_unflatten(treedef, leaves)

    @abstractmethod
    def __call__(self):
        """Return an updated state"""

    def setup(self):
        """Used in place of __init__"""

    def flatten(self):
        """Expose a default flatten method"""
        return jax.tree_util.tree_flatten(self)[0]


class Env(Obj):
    @abstractmethod
    def init(self):
        """Return an initialized state"""


class Agent(Obj):
    pass


deluca.field = field
deluca.Obj = Obj
deluca.Env = Env
deluca.Agent = Agent
deluca.save = save
deluca.load = load



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field
from deluca.core import Obj


class PendulumState(Obj):
    arr: jnp.ndarray = field(trainable=True)
    h: int = field(0, trainable=True)


class Pendulum(Env):
    m: float = field(1.0, trainable=False)
    l: float = field(1.0, trainable=False)
    g: float = field(9.81, trainable=False)
    max_torque: float = field(1.0, trainable=False)
    dt: float = field(0.02, trainable=False)
    H: int = field(300, trainable=False)
    goal_state: jnp.ndarray = field(jnp.array([0.0, -1.0, 0.0]), trainable=False)

    def init(self):
        return PendulumState(arr=jnp.array([0.0, 1.0, 0.0]))

    def __call__(self, state, action):
        sin, cos, thdot = state.arr
        action = self.max_torque * jnp.tanh(action[0])
        th = jnp.arctan2(sin, cos)
        newthdot = th + (
            -3.0 * self.g / (2.0 * self.l) * jnp.sin(th + jnp.pi)
            + 3.0 / (self.m * self.l ** 2) * action
        )
        newth = th + newthdot * self.dt
        newsin, newcos = jnp.sin(newth), jnp.cos(newth)
        arr = jnp.array([newsin, newcos, newthdot])
        return PendulumState(arr=arr, h=state.h + 1), arr



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from deluca.core import Agent
from deluca.core import field


class BangBang(Agent):
    target: float = field(0.0, trainable=False)
    min_action: float = field(0.0, trainable=False)
    max_action: float = field(100.0, trainable=False)

    def init(self):
        return None

    def __call__(self, state, obs):
        """Assume observation is env state"""
        return state, self.min_action if obs > self.target else self.max_action



#!/usr/bin/env python3
import argparse
import fileinput
import pathlib
import shutil
import sys

parser = argparse.ArgumentParser(
    description="A script that generates the template for a new plugin"
)
parser.add_argument("name", type=str, help="Name of the package, formatted as deluca-name")
parser.add_argument(
    "-d",
    "--dst",
    type=pathlib.Path,
    help="Parent directory to place new plugin (default: 'PATH/deluca/..')",
    default=pathlib.Path(__file__).absolute().parent.parent.parent,
)

args = parser.parse_args()


def err_and_die(message):
    print(f"{message}\n")
    parser.print_help()
    sys.exit(1)


if not args.dst.exists() or not args.dst.is_dir():
    err_and_die(f"ERROR: '{args.dst}' either does not exist or is not a directory!")

path = args.dst.absolute() / f"deluca-{args.name}"

if path.exists():
    err_and_die(f"ERROR: '{path}' already exists!")

src = pathlib.Path(__file__).absolute().parent.parent / "plugin"

shutil.copytree(src, path)

for file_path in path.glob("**/*"):
    if file_path.is_file():
        with fileinput.FileInput(file_path.absolute(), inplace=True) as file:
            for line in file:
                line = line.replace("deluca.plugin", f"deluca.{args.name}")
                line = line.replace("deluca-plugin", f"deluca-{args.name}")
                line = line.replace("deluca/plugin", f"deluca/{args.name}")
                print(line, end="")

lib_path = path / "deluca" / args.name
lib_path.mkdir()

with open(lib_path / "__init__.py", "w") as init_file:
    pass

with open(lib_path / "_version.py", "w") as version_file:
    version_file.write("__version__ = 0.0.1")



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field
from deluca.core import Obj


def dissipative(x, y, wind):
    return [wind * x, wind * y]


def constant(x, y, wind):
    return [wind * jnp.cos(jnp.pi / 4.0), wind * jnp.sin(jnp.pi / 4.0)]


class PlanarQuadrotorState(Obj):
    arr: jnp.ndarray = field(trainable=True)
    h: int = field(0, trainable=True)


class PlanarQuadrotor(Env):
    m: float = field(0.1, trainable=False)
    l: float = field(0.2, trainable=False)
    g: float = field(9.81, trainable=False)
    dt: float = field(0.05, trainable=False)
    H: int = field(100, trainable=False)
    wind: float = field(0.0, trainable=False)
    wind_func: Callable = field(dissipative, trainable=False)
    goal_state: jnp.ndarray = field(jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]), trainable=False)
    goal_action: jnp.ndarray = field(trainable=False)
    last_action: jnp.ndarray = field(jnp.array([0.0, 0.0]), trainable=False)

    def setup(self):
        self.goal_action = jnp.array([self.m * self.g / 2.0, self.m * self.g / 2.0])

    def _wind_field(self, x, y):
        return self.wind_func(x, y, self.wind)

    def init(self):
        return PlanarQuadrotorState(arr=jnp.array[1.0, 1.0, 0.0, 0.0, 0.0, 0.0], h=0)

    def __call__(self, state, action):
        x, y, th, xdot, ydot, thdot = state.arr
        u1, u2 = action
        m, g, l, dt = self.m, self.g, self.l, self.dt
        wind = self.wind_field(x, y)
        xddot = -(u1 + u2) * jnp.sin(th) / m + wind[0] / m
        yddot = (u1 + u2) * jnp.cos(th) / m - g + wind[1] / m
        thddot = l * (u2 - u1) / (m * l ** 2)
        state_dot = jnp.array([xdot, ydot, thdot, xddot, yddot, thddot])
        arr = state.arr + state_dot * dt
        return state.replace(arr=arr, last_action=action, h=state.h + 1), arr
